
// This file is part of NALib, a collection of C and C++ source code
// intended for didactical purposes. Full license notice at the bottom.

#include "NABinaryData.h"
#include "NARange.h"
#include <stdlib.h>







// ////////////////////////////////
// Endianness converter struct definitions
// ////////////////////////////////


NABool naIsEndiannessNative(NAInt endianness){
  #ifndef NDEBUG
    if((endianness < 0) || (endianness >= NA_ENDIANNESS_COUNT))
      naError("naIsEndiannessNative", "endianness is invalid. Result is undefined.");
  #endif
  return ( (endianness == NA_ENDIANNESS_NATIVE)
        || (endianness == NA_ENDIANNESS_UNKNOWN)
        || (endianness == NA_SYSTEM_ENDIANNESS));
}

void naInitEndiannessConverter(NAEndiannessConverter* converter, NAInt end1, NAInt end2){
  #ifndef NDEBUG
    if((end1 < 0) || (end1 >= NA_ENDIANNESS_COUNT))
      naError("createEndiannessConverter", "endianness1 is invalid. Result is undefined.");
    if((end2 < 0) || (end2 >= NA_ENDIANNESS_COUNT))
      naError("createEndiannessConverter", "endianness2 is invalid. Result is undefined.");
  #endif
  if((end1 == NA_ENDIANNESS_NATIVE) || (end1 == NA_ENDIANNESS_UNKNOWN)){
    end1 = NA_SYSTEM_ENDIANNESS;
  }
  if((end2 == NA_ENDIANNESS_NATIVE) || (end2 == NA_ENDIANNESS_UNKNOWN)){
    end2 = NA_SYSTEM_ENDIANNESS;
  }
  // Note that currently, as only Little and Big endianness is implemented,
  // there are only two possibilities: Endiannesses are the same or they are
  // little and big. If there are more endiannesses implemented in the futute,
  // this implementation will likely become a little more complicated.
  if(end1 == end2){
    converter->convert128       = &naConvertNative128;
    converter->convert64        = &naConvertNative64;
    converter->convert32        = &naConvertNative32;
    converter->convert16        = &naConvertNative16;
    converter->convert8         = &naConvertNative8;
    converter->convertArray128  = &naConvertArrayNative128;
    converter->convertArray64   = &naConvertArrayNative64;
    converter->convertArray32   = &naConvertArrayNative32;
    converter->convertArray16   = &naConvertArrayNative16;
    converter->convertArray8    = &naConvertArrayNative8;
  }else{
    converter->convert128       = &naConvertLittleBig128;
    converter->convert64        = &naConvertLittleBig64;
    converter->convert32        = &naConvertLittleBig32;
    converter->convert16        = &naConvertLittleBig16;
    converter->convert8         = &naConvertLittleBig8;
    converter->convertArray128  = &naConvertArrayLittleBig128;
    converter->convertArray64   = &naConvertArrayLittleBig64;
    converter->convertArray32   = &naConvertArrayLittleBig32;
    converter->convertArray16   = &naConvertArrayLittleBig16;
    converter->convertArray8    = &naConvertArrayLittleBig8;
  }
}



#define NA_NATIVE_ENDIANNESS_MARKER     0x42446462
#if NA_SYSTEM_ENDIANNESS == NA_ENDIANNESS_LITTLE
  #define NA_LITTLE_ENDIANNESS_MARKER   0x42446462
  #define NA_BIG_ENDIANNESS_MARKER      0x62644442
#elif NA_SYSTEM_ENDIANNESS == NA_ENDIANNESS_BIG
  #define NA_LITTLE_ENDIANNESS_MARKER   0x62644442
  #define NA_BIG_ENDIANNESS_MARKER      0x42446462
#endif


void naFillEndiannessMarker(NAByte marker[4], NAInt endianness){
  switch(endianness){
  case NA_ENDIANNESS_LITTLE:
    (*((uint32*)marker)) = NA_LITTLE_ENDIANNESS_MARKER;
    break;
  case NA_ENDIANNESS_BIG:
    (*((uint32*)marker)) = NA_BIG_ENDIANNESS_MARKER;
    break;
  default:
    (*((uint32*)marker)) = NA_NATIVE_ENDIANNESS_MARKER;
    break;
  }
}

NAInt naParseEndiannessMarker(const NAByte marker[4]){
  switch(*((uint32*)marker)){
  case NA_LITTLE_ENDIANNESS_MARKER:
    return NA_ENDIANNESS_LITTLE;
    break;
  case NA_BIG_ENDIANNESS_MARKER:
    return NA_ENDIANNESS_BIG;
    break;
  default:
    return NA_ENDIANNESS_UNKNOWN;
    break;
  }
}




// //////////////////////////////////
// NAByteArray
// //////////////////////////////////



NAByteArray* naCreateByteArray(NAByteArray* array){
  array = naAllocateIfNull(array, sizeof(NAByteArray));
  array->size = 0;
  return array;
}



NAByteArray* naCreateByteArrayWithSize(NAByteArray* array, NAInt size){
  #ifndef NDEBUG
    if(size < 0)
      naError("naCreateByteArrayWithSize", "size parameter is negative.");
  #endif
  if(!size){  // if size is zero
    array = naCreateByteArray(array);
  }else{
    array = naAllocateIfNull(array, sizeof(NAByteArray));
    array->storage = naCreatePointerWithSize(size);
    array->ptr = naGetPointerData(array->storage);
    array->size = size;
  }
  return array;
}



NAByteArray* naCreateByteArrayWithExistingBuffer( NAByteArray* array,
                                                       void* buffer,
                                                      NAInt size,
                                                      NABool takeownership){
  #ifndef NDEBUG
    if(size < 0)
      naError("naCreateByteArrayWithExistingBuffer", "size parameter is negative.");
    if(!buffer)
      naError("naCreateByteArrayWithExistingBuffer", "buffer parameter is NULL.");
  #endif
  if(!size){  // if size is zero
    array = naCreateByteArray(array);
  }else{
    array = naAllocateIfNull(array, sizeof(NAByteArray));
    array->storage = naCreatePointerWithExistingBuffer(buffer, takeownership);
    array->ptr = naGetPointerData(array->storage);
    array->size = size;
  }
  return array;
}


NAByteArray* naCreateByteArrayExtraction( NAByteArray* dstarray,
                                  const NAByteArray* srcarray,
                                         NAInt offset,
                                         NAInt size){
  NAByte* newptr;
  NAInt newsize;

  #ifndef NDEBUG
    if(!srcarray)
      {naCrash("naCreateByteArrayExtraction", "Source parameter is NULL."); return NA_NULL;}
  #endif

  dstarray = naAllocateIfNull(dstarray, sizeof(NAByteArray));
  // Note that dstarray may be equal to srcarray.

  // First, set the offset and compute the size with the new offset.
  newptr = srcarray->ptr + offset;
  newsize = naGetByteArraySize(srcarray) - offset;

  if(offset < 0){
    // If offset is negative
    newptr += naGetByteArraySize(srcarray);
    newsize -= naGetByteArraySize(srcarray);
  }

  if(newsize < 0){
    #ifndef NDEBUG
      naError("naCreateByteArrayExtraction", "Invalid offset leads to array overflow. Correcting to empty array.");
    #endif
    newsize = 0;
  }else if(newsize > naGetByteArraySize(srcarray)){
    #ifndef NDEBUG
      naError("naCreateByteArrayExtraction", "Invalid offset leads to array underflow. Correcting to empty array.");
    #endif
    newsize = 0;
  }else{
    // The offset is valid. Now, adjust the size.
    if(size < 0){ // negative size parameter
      newsize += size + 1;
      if(newsize < 0){
        // When the resulting size is smaller than 0, underflow.
        #ifndef NDEBUG
          naError("naCreateByteArrayExtraction", "Invalid size leads to array underflow. Correcting to empty array.");
        #endif
        newsize = 0;
      }
    }else{ // positive size parameter
      if(size <= newsize){
        newsize = size;
      }else{
        // When the desired size is bigger than the size available, overflow.
        #ifndef NDEBUG
          naError("naCreateByteArrayExtraction", "Invalid size leads to array overflow. Correcting to empty array.");
        #endif
        newsize = 0;
      }
    }
  }

  if(!newsize){
    // If the extraction results in an empty array...
    if(dstarray == srcarray){
      naClearByteArray(dstarray); // clear the old array.
      naCreateByteArray(dstarray);
    }else{
      naCreateByteArray(dstarray);
    }
  }else{
    // The resulting array has content!
    if(dstarray != srcarray){
      dstarray = naCreateByteArray(dstarray);
      dstarray->storage = naRetainPointer(srcarray->storage);
    }
    dstarray->ptr = newptr;
    dstarray->size = newsize;
  }
  
  return dstarray;
}



void naClearByteArray(NAByteArray* array){
  #ifndef NDEBUG
    if(!array)
      {naCrash("naClearByteArray", "array parameter is NULL."); return;}
  #endif
  if(naGetByteArraySize(array)){naReleasePointer(array->storage);}
}


void naDestroyByteArray(NAByteArray* array){
  naClearByteArray(array);
  free(array);
}



void naDecoupleByteArray(NAByteArray* array, NAInt addbytes){
  #ifndef NDEBUG
    if(!array)
      {naCrash("naDecoupleByteArray", "array parameter is NULL."); return;}
  #endif
  if(!naGetByteArraySize(array)){
    // Array is empty.
    if(addbytes){naCreateByteArrayWithSize(array, naAbsi(addbytes));}
    return;
  }else{
    // Note: Do not use realloc as ptr may point to a subset of NAPointer
    NAPointer* newstorage;
    if(addbytes > 0){
      newstorage = naCreatePointerWithSize(naGetByteArraySize(array) + addbytes);
      naCpyn(naGetPointerData(newstorage), array->ptr + addbytes, naGetByteArraySize(array));
      array->size += addbytes;
    }else{
      newstorage = naCreatePointerWithSize(naGetByteArraySize(array) - addbytes);
      naCpyn(naGetPointerData(newstorage), array->ptr, naGetByteArraySize(array));
      array->size -= addbytes;
    }
    naReleasePointer(array->storage);
    array->storage = newstorage;
    array->ptr = naGetPointerData(array->storage);
  }
}



NAByte* naGetByteArrayByte(NAByteArray* array, NAInt index){
  #ifndef NDEBUG
    if(!array)
      {naCrash("naGetByteArrayByte", "array parameter is NULL."); return NA_NULL;}
    if(!naGetByteArraySize(array))
      naError("naGetByteArrayByte", "array is empty");
  #endif
  if(index < 0){index += naGetByteArraySize(array);}
  #ifndef NDEBUG
    if(!naInsidei(index, 0, naGetByteArraySize(array)))
      naError("naGetByteArrayByte", "index out of bounds");
  #endif
  return &(array->ptr[index]);
}


NAInt naGetByteArraySize(const NAByteArray* array){
  #ifndef NDEBUG
    if(!array)
      {naCrash("naGetByteArraySize", "array parameter is NULL."); return 0;}
    if(array->size < 0)
      naError("naGetByteArraySize", "returned size is negative. Uninitialized struct?");
  #endif
  return array->size;
}










// Copyright (c) NALib, Tobias Stamm, Manderim GmbH
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the source-code.
//
// In case the source-code of this software is inaccessible to the end-user,
// the above copyright notice and this permission notice shall be included
// in any source-code which is dependent on this software and is accessible
// to the end-user.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
