
// This file is part of NALib, a collection of C and C++ source code
// intended for didactical purposes. Full license notice at the bottom.

#ifndef NA_BINARY_DATA_INCLUDED
#define NA_BINARY_DATA_INCLUDED

#include "NASystem.h"
#include "NAPointer.h"
#include <string.h>

// This file provides basic functionality to handle binary data.
// You can use the functions provided here directly. Often times though you
// can just use the NAByteArray struct defined in the appropriate file.
//
// The following functions are available for 8, 16, 32, 64 and 128 Bits:
// - assigning byte values
// - comparing byte values
// - swapping of multi-byte-values
// - Endianness conversion
// - Endianness helper functions



// ////////////////
// Fills the array of the first parameter with the given bytes in increasing
// address positions.
// ////////////////

static void naSet128  (void* a,
                      NAByte b0,  NAByte b1,  NAByte b2,  NAByte b3,
                      NAByte b4,  NAByte b5,  NAByte b6,  NAByte b7,
                      NAByte b8,  NAByte b9,  NAByte b10, NAByte b11,
                      NAByte b12, NAByte b13, NAByte b14, NAByte b15);
static void naSet64   (void* a,
                      NAByte b0,  NAByte b1,  NAByte b2,  NAByte b3,
                      NAByte b4,  NAByte b5,  NAByte b6,  NAByte b7);
static void naSet32   (void* a,
                      NAByte b0,  NAByte b1,  NAByte b2,  NAByte b3);
static void naSet16   (void* a,
                      NAByte b0,  NAByte b1);
static void naSet8    (void* a,
                      NAByte b0);


// ////////////////
// Copies the array b to array a byte by byte
// ////////////////

static void naCpyn    (void* a, const void* b, NAInt count);
static void naCpy128  (void* a, const void* b);
static void naCpy64   (void* a, const void* b);
static void naCpy32   (void* a, const void* b);
static void naCpy16   (void* a, const void* b);
static void naCpy8    (void* a, const void* b);


// ////////////////
// Sets the given array a to binary 0
// ////////////////

static void naNulln   (void* a, NAInt count);
static void naNull128 (void* a);
static void naNull64  (void* a);
static void naNull32  (void* a);
static void naNull16  (void* a);
static void naNull8   (void* a);

// Fills the given array with Bytes of the value 1
static void naOnen    (void* a, NAInt count);


// ////////////////
// Compares the array of the first parameter with the given bytes in increasing
// address positions and returns NA_TRUE only if they are all equal.
// ////////////////

static NABool naEqual128 (const void* a,
                               NAByte b0,  NAByte b1,  NAByte b2,  NAByte b3,
                               NAByte b4,  NAByte b5,  NAByte b6,  NAByte b7,
                               NAByte b8,  NAByte b9,  NAByte b10, NAByte b11,
                               NAByte b12, NAByte b13, NAByte b14, NAByte b15);
static NABool naEqual64  (const void* a,
                               NAByte b0,  NAByte b1,  NAByte b2,  NAByte b3,
                               NAByte b4,  NAByte b5,  NAByte b6,  NAByte b7);
static NABool naEqual32  (const void* a,
                               NAByte b0,  NAByte b1,  NAByte b2,  NAByte b3);
static NABool naEqual16  (const void* a,
                               NAByte b0,  NAByte b1);
static NABool naEqual8   (const void* a,
                               NAByte b0);



// ////////////////
// Swaps the content of the two pointers with the given number of bits.
//
// Note: These functions are implemented with a well known swapping trick which
// only works if the two pointers do not point to the same location. Therefore
// restrict pointers make prefectly sense here.
// ////////////////

static void naSwap128 (void* NA_RESTRICT a, void* NA_RESTRICT b);
static void naSwap64  (void* NA_RESTRICT a, void* NA_RESTRICT b);
static void naSwap32  (void* NA_RESTRICT a, void* NA_RESTRICT b);
static void naSwap16  (void* NA_RESTRICT a, void* NA_RESTRICT b);
static void naSwap8   (void* NA_RESTRICT a, void* NA_RESTRICT b);



// ////////////////
// Static Endianness conversions.
// Use those if you know exactly from which endianness to which you want to
// convert. If you don't know it at compile-time, use the NAEndiannessConverter
// below or NAByteArray.
//
// Will convert the data directly within the buffer
// provided. All converters work in both ways, for example Little to Big
// will also convert Big to Little.
//
// The Array-variants convert a whole array with count elements. The
// non-Array-variants convert only one single value.
// ////////////////

// Conversion between two same Endiannesses.
static  void naConvertNative128(            void* buffer);
static  void naConvertNative64(             void* buffer);
static  void naConvertNative32(             void* buffer);
static  void naConvertNative16(             void* buffer);
static  void naConvertNative8(              void* buffer);
static  void naConvertArrayNative128(       void* buffer, NAInt count);
static  void naConvertArrayNative64(        void* buffer, NAInt count);
static  void naConvertArrayNative32(        void* buffer, NAInt count);
static  void naConvertArrayNative16(        void* buffer, NAInt count);
static  void naConvertArrayNative8(         void* buffer, NAInt count);

// Conversion between Little and Big Endianness
static  void naConvertLittleBig128(         void* buffer);
static  void naConvertLittleBig64(          void* buffer);
static  void naConvertLittleBig32(          void* buffer);
static  void naConvertLittleBig16(          void* buffer);
static  void naConvertLittleBig8(           void* buffer);
static  void naConvertArrayLittleBig128(    void* buffer, NAInt count);
static  void naConvertArrayLittleBig64(     void* buffer, NAInt count);
static  void naConvertArrayLittleBig32(     void* buffer, NAInt count);
static  void naConvertArrayLittleBig16(     void* buffer, NAInt count);
static  void naConvertArrayLittleBig8(      void* buffer, NAInt count);

// Conversion between Native and Little Endianness
static  void naConvertNativeLittle128(      void* buffer);
static  void naConvertNativeLittle64(       void* buffer);
static  void naConvertNativeLittle32(       void* buffer);
static  void naConvertNativeLittle16(       void* buffer);
static  void naConvertNativeLittle8(        void* buffer);
static  void naConvertArrayNativeLittle128( void* buffer, NAInt count);
static  void naConvertArrayNativeLittle64(  void* buffer, NAInt count);
static  void naConvertArrayNativeLittle32(  void* buffer, NAInt count);
static  void naConvertArrayNativeLittle16(  void* buffer, NAInt count);
static  void naConvertArrayNativeLittle8(   void* buffer, NAInt count);

// Conversion between Native and Big Endianness
static  void naConvertNativeBig128(         void* buffer);
static  void naConvertNativeBig64(          void* buffer);
static  void naConvertNativeBig32(          void* buffer);
static  void naConvertNativeBig16(          void* buffer);
static  void naConvertNativeBig8(           void* buffer);
static  void naConvertArrayNativeBig128(    void* buffer, NAInt count);
static  void naConvertArrayNativeBig64(     void* buffer, NAInt count);
static  void naConvertArrayNativeBig32(     void* buffer, NAInt count);
static  void naConvertArrayNativeBig16(     void* buffer, NAInt count);
static  void naConvertArrayNativeBig8(      void* buffer, NAInt count);



// ////////////////
// If you have dynamically defined endiannesses, use an NAEndiannessConverter.
// Use the following function to initialize the structure. After that, you can
// use the function-pointers of the NAEndiannessConverter struct directly.
// ////////////////

typedef struct NAEndiannessConverter NAEndiannessConverter;
struct NAEndiannessConverter{
  void (*convert128)      (void* buffer);
  void (*convert64)       (void* buffer);
  void (*convert32)       (void* buffer);
  void (*convert16)       (void* buffer);
  void (*convert8)        (void* buffer);
  void (*convertArray128) (void* buffer, NAInt count);
  void (*convertArray64)  (void* buffer, NAInt count);
  void (*convertArray32)  (void* buffer, NAInt count);
  void (*convertArray16)  (void* buffer, NAInt count);
  void (*convertArray8)   (void* buffer, NAInt count);
};

void naInitEndiannessConverter  (NAEndiannessConverter* converter,
                                                  NAInt endianness1,
                                                  NAInt endianness2);


// //////////////////////////////////
// naIsEndiannessNative returns NA_TRUE if the given endianness is the same as
// the endianness of this system. Also returns NA_TRUE when sending native or
// unknown endianness. This function is useful for dynamically checking for
// endiannesses. If you simply are interested in the (static) endianness of
// this system, check the NA_SYSTEM_ENDIANNESS macro.
// //////////////////////////////////

NABool naIsEndiannessNative(NAInt endianness);


// ////////////////////////////////
// Use the following functions to write and read a 4-byte endianness
// marker. This is useful for files which state what endianness they
// are encoded with. Important notice: Read and write the marker as a
// raw 4-byte-array, not as a 4-byte-entity!
//
// The markers are
// - "BDdb" for Big Endian
// - "bdDB" for Little Endian
// Note that the marker will not likely change in the future.
// ///////////////////////////////

void  naFillEndiannessMarker (NAByte marker[4], NAInt endianness);
NAInt naParseEndiannessMarker(const NAByte marker[4]);






// /////////////////////////
// NAByteArray
//
// An NAByteArray stores an array of raw bytes by dereferencing a storage.
// Every ByteArray has a pointer to an internal storage, a data pointer and a
// size.
//
// Multiple NAByteArrays can point to the same storage. Meaning: Multiple
// NAByteArrays can point to different sections of the same byte stream by
// providing an offset and a size relative to another NAByteArray. This allows
// for very fast and convenient parsing of byte streams without copying
// the bytes over and over again. This struct is also used by NAString and
// NABitArray
// /////////////////////////

typedef struct NAByteArray NAByteArray;
struct NAByteArray{
  NAPointer* storage;   // pointer to the storage of all bytes
  NAByte* ptr;          // pointer to the first byte of this NAByteArray
  NAInt size;           // size of this NAByteArray in bytes
};
// Note that ptr may not point to the same address as the data pointer stored
// in the storage!
//
// If size is zero, the ByteArray is considered empty. In that case, the
// other fields contain no useful information.



// ///////////////////////////
// Constructors and Destructors:
// Search for CONSTRUCTORS and DESTRUCTORS in NAPointer.h for more information.
// ///////////////////////////

// Creates or fills a new, EMPTY ByteArray and returns the pointer to the
// NAByteArray struct. Other than the struct itself, no additional memory will
// be allocated.
NAByteArray* naCreateByteArray(NAByteArray* array);

// Creates or fills a new ByteArray with the desired size and returns the
// pointer. Allocates memory on the heap which will be freed automatically when
// the array storage is no longer used. size shall not be negative. If size
// is zero, an empty array is returned.
NAByteArray* naCreateByteArrayWithSize(NAByteArray* array,
                                              NAInt size);

// Creates or fills a new ByteArray which contains the data of the given
// buffer WITHOUT copying. The size denotes the size of the buffer in bytes.
// The programmer is responsible for that size does not overflows the buffer.
// When takeownership is set to NA_TRUE, the array will deallocate the given
// buffer with free() when it is no longer used.
NAByteArray* naCreateByteArrayWithExistingBuffer( NAByteArray* array,
                                                         void* buffer,
                                                         NAInt size,
                                                        NABool takeownership);

// Function naCreateByteArrayExtraction:
// Creates or fills a new ByteArray with the contents of srcarray. The content
// is not copied, but the dest array points to the same storage as srcarray
// with an offset and size relative to the src array.
//
// Use this function to perform all sorts of manipulations (examples below):
// - dstarray can be the same as srcarray
// - if offset is negative, it denotes the number of bytes from the end.
//   Note that the end has index [size], meaning -1 denotes the index [size-1].
// - If the size is 0, the resulting ByteArray is empty.
// - if size is negative, it denotes the size up and including to the given
//   number of bytes from the end, meaning -1 denotes the last byte.
// - if the offset and size combination somehow leads to a size of exactly 0,
//   the resulting ByteArray will be empty without a warning emitted.
// - If the offset and size combination somehow leads to an over- or underflow,
//   a warning will be emitted if NDEBUG is defined. The resulting array will
//   be empty.
//
// Example: Array ABCDEF with the pair (offset, size):
// - ( 2,  2)   ->   CD   (extraction)
// - ( 2,  0)   ->   CDEF (truncate head from left)
// - (-2,  0)   ->     EF (truncate head from right)
// - ( 0,  2)   -> AB     (truncate tail from left)
// - ( 0, -3)   -> ABCD   (truncate tail from right)
// - (-3, -2)   ->    DE  (mix it as you desire)
// - ( 0, -1)   -> ABCDEF (exact duplicate)
// - ( 1,  0)   ->        (empty array because of desired size = 0)
// - ( 2, -5)   ->        (empty array because of resulting size = 0)
// - (-9,  9)   ->        (empty array with warning when debugging)
//
// If dest ist equal to src and if the resulting array is empty, the storage
// is automatically detached and freed if no longer used.
//
// Warning: srcarray may be const but if dstarray == srcarray, srcarray may
// nontheless be altered!!!
NAByteArray* naCreateByteArrayExtraction( NAByteArray* dstarray,
                                    const NAByteArray* srcarray,
                                                 NAInt offset,
                                                 NAInt size);

// Clears or Deletes the given array. Any previously used storage is detached.
// When the storage is no longer used, it gets deleted automatically.
void naClearByteArray(NAByteArray* array);
void naDestroyByteArray(NAByteArray* array);

// COPIES the contents of the array to a separate storage and decouples it
// from the existing storage. The additionalbytes are added if requested
// and will be uninitialized.
//
// When additionalbytes is positive, the bytes get added at the beginning
// of the array, when negative, at the end. Note that this is in accordance
// with other parameters used in Array structs where positive numbers denote
// the beginning and negative numbers the end. Here, in most cases though, a
// programmer wants to add at the end, so this parameter is likely negative.
// If this parameter is 0, nothing is appended.
//
// When the existing storage is no longer used, it gets deleted automatically.
//
// COPIES ALWAYS!
void naDecoupleByteArray(NAByteArray* array, NAInt addbytes);

// Returns a POINTER to the byte at the given index. If index is negative, it
// is searched from the end of the array. For example -1 denotes the last byte.
// Note that calling this function too often might not be a good idea with
// respect to speed. Try getting one byte using this function and use pointer
// arithmetic afterwards.
NAByte* naGetByteArrayByte(NAByteArray* array, NAInt index);

// Returns the number of bytes in this array. If you need this value in a loop,
// try storing it in a variable.
NAInt naGetByteArraySize(const NAByteArray* array);







// ////////////////////////////////////////////////////////////////////
// ////////////////////////////////////////////////////////////////////
//
// End of API. Everything below this point is implementation and details.
//
// ////////////////////////////////////////////////////////////////////
// ////////////////////////////////////////////////////////////////////






// ///////////////////////////////////////////////////////////
// Fill multiple bytes into the content of a given pointer
// ///////////////////////////////////////////////////////////

static NA_INLINE void naSet128( void* a,
                            NAByte b0,  NAByte b1,  NAByte b2,  NAByte b3,
                            NAByte b4,  NAByte b5,  NAByte b6,  NAByte b7,
                            NAByte b8,  NAByte b9,  NAByte b10, NAByte b11,
                            NAByte b12, NAByte b13, NAByte b14, NAByte b15){
  register NAByte* p; // Declaration befor implementation;
  #ifndef NDEBUG
    if(!a)
      {naCrash("naSet128", "Pointer is NULL."); return;}
  #endif
  p = (NAByte*)a;
  *p++ = b0;  *p++ = b1;  *p++ = b2;  *p++ = b3;
  *p++ = b4;  *p++ = b5;  *p++ = b6;  *p++ = b7;
  *p++ = b8;  *p++ = b9;  *p++ = b10; *p++ = b11;
  *p++ = b12; *p++ = b13; *p++ = b14; *p   = b15;
}

static NA_INLINE void naSet64( void* a,
                           NAByte b0, NAByte b1, NAByte b2, NAByte b3,
                           NAByte b4, NAByte b5, NAByte b6, NAByte b7){
  register NAByte* p; // Declaration befor implementation;
  #ifndef NDEBUG
    if(!a)
      {naCrash("naSet64", "Pointer is NULL."); return;}
  #endif
  p = (NAByte*)a;
  *p++ = b0; *p++ = b1; *p++ = b2; *p++ = b3;
  *p++ = b4; *p++ = b5; *p++ = b6; *p   = b7;
}

static NA_INLINE void naSet32( void* a,
                           NAByte b0, NAByte b1, NAByte b2, NAByte b3){
  register NAByte* p; // Declaration befor implementation;
  #ifndef NDEBUG
    if(!a)
      {naCrash("naSet32", "Pointer is NULL."); return;}
  #endif
  p = (NAByte*)a;
  *p++ = b0; *p++ = b1; *p++ = b2; *p = b3;
}

static NA_INLINE void naSet16( void* a,
                           NAByte b0, NAByte b1){
  register NAByte* p; // Declaration befor implementation;
  #ifndef NDEBUG
    if(!a)
      {naCrash("naSet16", "Pointer is NULL."); return;}
  #endif
  p = (NAByte*)a;
  *p++ = b0; *p = b1;
}

static NA_INLINE void naSet8( void* a,
                          NAByte b0){
  register NAByte* p; // Declaration befor implementation;
  #ifndef NDEBUG
    if(!a)
      {naCrash("naSet8", "Pointer is NULL."); return;}
  #endif
  p = (NAByte*)a;
  *p = b0;
}


// ///////////////////////////////////////////////////////////
// Copy multiple bytes from b to a
// ///////////////////////////////////////////////////////////

static void naCpyn(void* a, const void* b, NAInt count){
  memcpy(a, b, count);
}
static void naCpy128(void* a, const void* b){
  *(uint64*)a = *(uint64*)b;
  a = ((NAByte*)a) + 8;
  b = ((NAByte*)b) + 8;
  *(uint64*)a = *(uint64*)b;
}
static void naCpy64 (void* a, const void* b){
  *(uint64*)a = *(uint64*)b;
}
static void naCpy32 (void* a, const void* b){
  *(uint32*)a = *(uint32*)b;
}
static void naCpy16 (void* a, const void* b){
  *(uint16*)a = *(uint16*)b;
}
static void naCpy8  (void* a, const void* b){
  *(uint8*)a = *(uint8*)b;
}


// ///////////////////////////////////////////////////////////
// Nullifies the array
// ///////////////////////////////////////////////////////////

static void naNulln(void* a, NAInt count){
  // Note that the bzero function does the same but is deprecated.
  memset(a, 0, count);
}
static void naNull128(void* a){
  *(uint64*)a = 0LL;
  a = ((NAByte*)a) + 8;
  *(uint64*)a = 0LL;
}
static void naNull64 (void* a){
  *(uint64*)a = 0LL;
}
static void naNull32 (void* a){
  *(uint32*)a = 0;
}
static void naNull16 (void* a){
  *(uint16*)a = 0;
}
static void naNull8  (void* a){
  *(uint8*)a = 0;
}


static void naOnen(void* a, NAInt count){
  memset(a, 0x01, count);
}

// ///////////////////////////////////////////////////////////
// Compare multiple bytes to the contents of a given pointer
// ///////////////////////////////////////////////////////////

static NA_INLINE NABool naEqual128(const void* a,
                              NAByte b0,  NAByte b1,  NAByte b2,  NAByte b3,
                              NAByte b4,  NAByte b5,  NAByte b6,  NAByte b7,
                              NAByte b8,  NAByte b9,  NAByte b10, NAByte b11,
                              NAByte b12, NAByte b13, NAByte b14, NAByte b15){
  register NAByte* p; // Declaration before implementation
  #ifndef NDEBUG
    if(!a)
      {naCrash("naEqual128", "Pointer is NULL."); return NA_FALSE;}
  #endif
  p = (NAByte*)a;
  if(*p++ != b0) {return NA_FALSE;}
  if(*p++ != b1) {return NA_FALSE;}
  if(*p++ != b2) {return NA_FALSE;}
  if(*p++ != b3) {return NA_FALSE;}
  if(*p++ != b4) {return NA_FALSE;}
  if(*p++ != b5) {return NA_FALSE;}
  if(*p++ != b6) {return NA_FALSE;}
  if(*p++ != b7) {return NA_FALSE;}
  if(*p++ != b8) {return NA_FALSE;}
  if(*p++ != b9) {return NA_FALSE;}
  if(*p++ != b10){return NA_FALSE;}
  if(*p++ != b11){return NA_FALSE;}
  if(*p++ != b12){return NA_FALSE;}
  if(*p++ != b13){return NA_FALSE;}
  if(*p++ != b14){return NA_FALSE;}
  if(*p   != b15){return NA_FALSE;}
  return NA_TRUE;
}

static NA_INLINE NABool naEqual64(const void* a,
                               NAByte b0, NAByte b1, NAByte b2, NAByte b3,
                               NAByte b4, NAByte b5, NAByte b6, NAByte b7){
  register NAByte* p; // Declaration before implementation
  #ifndef NDEBUG
    if(!a)
      {naCrash("naEqual64", "Pointer is NULL."); return NA_FALSE;}
  #endif
  p = (NAByte*)a;
  if(*p++ != b0){return NA_FALSE;}
  if(*p++ != b1){return NA_FALSE;}
  if(*p++ != b2){return NA_FALSE;}
  if(*p++ != b3){return NA_FALSE;}
  if(*p++ != b4){return NA_FALSE;}
  if(*p++ != b5){return NA_FALSE;}
  if(*p++ != b6){return NA_FALSE;}
  if(*p   != b7){return NA_FALSE;}
  return NA_TRUE;
}

static NA_INLINE NABool naEqual32(const void* a,
                               NAByte b0, NAByte b1, NAByte b2, NAByte b3){
  register NAByte* p; // Declaration before implementation
  #ifndef NDEBUG
    if(!a)
      {naCrash("naEqual32", "Pointer is NULL."); return NA_FALSE;}
  #endif
  p = (NAByte*)a;
  if(*p++ != b0){return NA_FALSE;}
  if(*p++ != b1){return NA_FALSE;}
  if(*p++ != b2){return NA_FALSE;}
  if(*p   != b3){return NA_FALSE;}
  return NA_TRUE;
}

static NA_INLINE NABool naEqual16(const void* a, NAByte b0, NAByte b1){
  register NAByte* p; // Declaration before implementation
  #ifndef NDEBUG
    if(!a)
      {naCrash("naEqual16", "Pointer is NULL."); return NA_FALSE;}
  #endif
  p = (NAByte*)a;
  if(*p++ != b0){return NA_FALSE;}
  if(*p   != b1){return NA_FALSE;}
  return NA_TRUE;
}

static NA_INLINE NABool naEqual8(const void* a, NAByte b0){
  register NAByte* p; // Declaration before implementation
  #ifndef NDEBUG
    if(!a)
      {naCrash("naEqual8", "Pointer is NULL."); return NA_FALSE;}
  #endif
  p = (NAByte*)a;
  if(*p   != b0){return NA_FALSE;}
  return NA_TRUE;
}




// /////////////////////////////
// Swap multibyte-values
// /////////////////////////////

static NA_INLINE void naSwap128(void* NA_RESTRICT a, void* NA_RESTRICT b){
  #ifndef NDEBUG
    NAInt dist = (NAByte*)a-(NAByte*)b;
    if((NAByte*)a<(NAByte*)b){dist = -dist;};
    if(dist < 16)
      naError("naSwap128", "Restrict pointers overlap.");
    if(!a)
      {naCrash("naSwap128", "Pointer a is NULL"); return;}
    if(!b)
      {naCrash("naSwap128", "Pointer b is NULL"); return;}
  #endif
  naSwap64(a, b);
  a = ((NAByte*)a) + 8;
  b = ((NAByte*)b) + 8;
  naSwap64(a, b);
}

static NA_INLINE void naSwap64(void* NA_RESTRICT a, void* NA_RESTRICT b){
  #ifndef NDEBUG
    NAInt dist = (NAByte*)a-(NAByte*)b;
    if((NAByte*)a<(NAByte*)b){dist = -dist;};
    if(dist < 8)
      naError("naSwap64", "Restrict pointers overlap.");
    if(!a)
      {naCrash("naSwap64", "Pointer a is NULL"); return;}
    if(!b)
      {naCrash("naSwap64", "Pointer b is NULL"); return;}
  #endif
  // Note: Do not write the following 3 lines as 1 line. The compiler might
  // cache the result of the dereference operators!
  *(uint64*)a^=*(uint64*)b;
  *(uint64*)b^=*(uint64*)a;
  *(uint64*)a^=*(uint64*)b;
}

static NA_INLINE void naSwap32(void* NA_RESTRICT a, void* NA_RESTRICT b){
  #ifndef NDEBUG
    NAInt dist = (NAByte*)a-(NAByte*)b;
    if((NAByte*)a<(NAByte*)b){dist = -dist;};
    if(dist < 4)
      naError("naSwap32", "Restrict pointers overlap.");
    if(!a)
      {naCrash("naSwap32", "Pointer a is NULL"); return;}
    if(!b)
      {naCrash("naSwap32", "Pointer b is NULL"); return;}
  #endif
  // Note: Do not write the following 3 lines as 1 line. The compiler might
  // cache the result of the dereference operators!
  *(uint32*)a^=*(uint32*)b;
  *(uint32*)b^=*(uint32*)a;
  *(uint32*)a^=*(uint32*)b;
}

static NA_INLINE void naSwap16(void* NA_RESTRICT a, void* NA_RESTRICT b){
  #ifndef NDEBUG
    NAInt dist = (NAByte*)a-(NAByte*)b;
    if((NAByte*)a<(NAByte*)b){dist = -dist;};
    if(dist < 2)
      naError("naSwap16", "Restrict pointers overlap.");
    if(!a)
      {naCrash("naSwap16", "Pointer a is NULL"); return;}
    if(!b)
      {naCrash("naSwap16", "Pointer b is NULL"); return;}
  #endif
  // Note: Do not write the following 3 lines as 1 line. The compiler might
  // cache the result of the dereference operators!
  *(uint16*)a^=*(uint16*)b;
  *(uint16*)b^=*(uint16*)a;
  *(uint16*)a^=*(uint16*)b;
}

static NA_INLINE void naSwap8(void* NA_RESTRICT a, void* NA_RESTRICT b){
  #ifndef NDEBUG
    NAInt dist = (NAByte*)a-(NAByte*)b;
    if((NAByte*)a<(NAByte*)b){dist = -dist;};
    if(dist < 1)
      naError("naSwap8", "Restrict pointers overlap.");
    if(!a)
      {naCrash("naSwap8", "Pointer a is NULL"); return;}
    if(!b)
      {naCrash("naSwap8", "Pointer b is NULL"); return;}
  #endif
  // Note: Do not write the following 3 lines as 1 line. The compiler might
  // cache the result of the dereference operators!
  *(uint8*)a^=*(uint8*)b;
  *(uint8*)b^=*(uint8*)a;
  *(uint8*)a^=*(uint8*)b;
}



// //////////////////////////////////////////////////////////////////////
// Endianness conversions
//
// The Endianness-Conversion functions are divided into single-value-converters
// and array-converters. The Array converters simply go through a whole array
// of single values and convert them using the corresponding single-value-
// converter.
//
// Note that the array-converters do not contain a test if count == 0 when
// debugging, as it will never result in an error. Even more so, there are
// situations, where sending a count of 0 is convenient.
// //////////////////////////////////////////////////////////////////////


// //////////////////////////////////////////////////////////////////////
// Convert between two same Endiannesses. Single converters
//
// Nothing needs to be converted. The buffer = buffer forces the compiler
// to not emit warnings.
// //////////////////////////////////////////////////////////////////////

static NA_INLINE void naConvertNative128(void* buffer){
  #ifndef NDEBUG
    if(!buffer)
      naError("naConvertN128", "Buffer is NULL");
  #endif
  buffer = buffer;
}

static NA_INLINE void naConvertNative64(void* buffer){
  #ifndef NDEBUG
    if(!buffer)
      naError("naConvertN64", "Buffer is NULL");
  #endif
  buffer = buffer;
}

static NA_INLINE void naConvertNative32(void* buffer){
  #ifndef NDEBUG
    if(!buffer)
      naError("naConvertN32", "Buffer is NULL");
  #endif
  buffer = buffer;
}

static NA_INLINE void naConvertNative16(void* buffer){
  #ifndef NDEBUG
    if(!buffer)
      naError("naConvertN16", "Buffer is NULL");
  #endif
  buffer = buffer;
}

static NA_INLINE void naConvertNative8(void* buffer){
  #ifndef NDEBUG
    if(!buffer)
      naError("naConvertN8", "Buffer is NULL");
  #endif
  buffer = buffer;
}


// //////////////////////////////
// Convert between two same Endiannesses. Array converters
//
// Nothing needs to be converted. The buffer=buffer and count=count forces the
// compiler to not emit warnings.
// //////////////////////////////

static NA_INLINE void naConvertArrayNative128(void* buffer, NAInt count){
  #ifndef NDEBUG
    if(count < 0)
      naError("naConvertArrayN128", "count is negative");
    if(!buffer)
      naError("naConvertArrayN128", "buffer is NULL");
  #endif
  buffer=buffer; count=count;
}

static NA_INLINE void naConvertArrayNative64(void* buffer, NAInt count){
  #ifndef NDEBUG
    if(count < 0)
      naError("naConvertArrayN64", "count is negative");
    if(!buffer)
      naError("naConvertArrayN64", "buffer is NULL");
  #endif
  buffer=buffer; count=count;
}

static NA_INLINE void naConvertArrayNative32(void* buffer, NAInt count){
  #ifndef NDEBUG
    if(count < 0)
      naError("naConvertArrayN32", "count is negative");
    if(!buffer)
      naError("naConvertArrayN32", "buffer is NULL");
  #endif
  buffer=buffer; count=count;
}

static NA_INLINE void naConvertArrayNative16(void* buffer, NAInt count){
  #ifndef NDEBUG
    if(count < 0)
      naError("naConvertArrayN16", "count is negative");
    if(!buffer)
      naError("naConvertArrayN16", "buffer is NULL");
  #endif
  buffer=buffer; count=count;
}

static NA_INLINE void naConvertArrayNative8(void* buffer, NAInt count){
  #ifndef NDEBUG
    if(count < 0)
      naError("naConvertArrayN8", "count is negative");
    if(!buffer)
      naError("naConvertArrayN8", "buffer is NULL");
  #endif
  buffer=buffer; count=count;
}



// //////////////////////////////////////////////////////////////////////
// Conversions Little to Big and vice versa. Single converters
// //////////////////////////////////////////////////////////////////////

static NA_INLINE void naConvertLittleBig128(void* buffer){
  #ifndef NDEBUG
    if(!buffer)
      {naCrash("naConvertLittleBig128", "Buffer is NULL"); return;}
  #endif
  naSwap8(((NAByte*)buffer)+0, ((NAByte*)buffer)+15);
  naSwap8(((NAByte*)buffer)+1, ((NAByte*)buffer)+14);
  naSwap8(((NAByte*)buffer)+2, ((NAByte*)buffer)+13);
  naSwap8(((NAByte*)buffer)+3, ((NAByte*)buffer)+12);
  naSwap8(((NAByte*)buffer)+4, ((NAByte*)buffer)+11);
  naSwap8(((NAByte*)buffer)+5, ((NAByte*)buffer)+10);
  naSwap8(((NAByte*)buffer)+6, ((NAByte*)buffer)+ 9);
  naSwap8(((NAByte*)buffer)+7, ((NAByte*)buffer)+ 8);
}

static NA_INLINE void naConvertLittleBig64(void* buffer){
  #ifndef NDEBUG
    if(!buffer)
      {naCrash("naConvertLittleBig64", "Buffer is NULL"); return;}
  #endif
  naSwap8(((NAByte*)buffer)+0, ((NAByte*)buffer)+7);
  naSwap8(((NAByte*)buffer)+1, ((NAByte*)buffer)+6);
  naSwap8(((NAByte*)buffer)+2, ((NAByte*)buffer)+5);
  naSwap8(((NAByte*)buffer)+3, ((NAByte*)buffer)+4);
}

static NA_INLINE void naConvertLittleBig32(void* buffer){
  #ifndef NDEBUG
    if(!buffer)
      {naCrash("naConvertLittleBig32", "Buffer is NULL"); return;}
  #endif
  naSwap8(((NAByte*)buffer)+0, ((NAByte*)buffer)+3);
  naSwap8(((NAByte*)buffer)+1, ((NAByte*)buffer)+2);
}

static NA_INLINE void naConvertLittleBig16(void* buffer){
  #ifndef NDEBUG
    if(!buffer)
      {naCrash("naConvertLittleBig16", "Buffer is NULL"); return;}
  #endif
  naSwap8(((NAByte*)buffer)+0, ((NAByte*)buffer)+1);
}

static NA_INLINE void naConvertLittleBig8(void* buffer){
  #ifndef NDEBUG
    if(!buffer)
      naError("naConvertLittleBig8", "Buffer is NULL");
  #endif
  // Bytes don't need to be converted. The following line forces the compiler
  // to not emit any warning.
  buffer = buffer;
}



// //////////////////////////////
// Conversions Little to Big and vice versa. Array converters
// //////////////////////////////

static NA_INLINE void naConvertArrayLittleBig128(void* buffer, NAInt count){
  NAInt i; // Declaration before implementation
  #ifndef NDEBUG
    if(count < 0)
      naError("naConvertArrayLittleBig128", "count is negative");
    if(!buffer)
      naError("naConvertArrayLittleBig128", "buffer is NULL");
  #endif
  for(i=0; i<count; i++){
    naConvertLittleBig128(buffer);
    buffer = (NAByte*)buffer + 16;
  }
}

static NA_INLINE void naConvertArrayLittleBig64(void* buffer, NAInt count){
  NAInt i; // Declaration before implementation
  #ifndef NDEBUG
    if(count < 0)
      naError("naConvertArrayLittleBig64", "count is negative");
    if(!buffer)
      naError("naConvertArrayLittleBig64", "buffer is NULL");
  #endif
  for(i=0; i<count; i++){
    naConvertLittleBig64(buffer);
    buffer = (NAByte*)buffer + 8;
  }
}

static NA_INLINE void naConvertArrayLittleBig32(void* buffer, NAInt count){
  NAInt i; // Declaration before implementation
  #ifndef NDEBUG
    if(count < 0)
      naError("naConvertArrayLittleBig32", "count is negative");
    if(!buffer)
      naError("naConvertArrayLittleBig32", "buffer is NULL");
  #endif
  for(i=0; i<count; i++){
    naConvertLittleBig32(buffer);
    buffer = (NAByte*)buffer + 4;
  }
}

static NA_INLINE void naConvertArrayLittleBig16(void* buffer, NAInt count){
  NAInt i; // Declaration before implementation
  #ifndef NDEBUG
    if(count < 0)
      naError("naConvertArrayLittleBig16", "count is negative");
    if(!buffer)
      naError("naConvertArrayLittleBig16", "buffer is NULL");
  #endif
  for(i=0; i<count; i++){
    naConvertLittleBig16(buffer);
    buffer = (NAByte*)buffer + 2;
  }
}

static NA_INLINE void naConvertArrayLittleBig8(void* buffer, NAInt count){
  #ifndef NDEBUG
    if(count < 0)
      naError("naConvertArrayLittleBig8", "count is negative");
    if(!buffer)
      naError("naConvertArrayLittleBig8", "buffer is NULL");
  #endif
  // Bytes don't need to be converted. The following line forces the compiler
  // to not emit any warning.
  buffer=buffer; count=count;
}




// //////////////////////////////////////
// All other endianness converters are simply redirects to the already defined
// converters, depending on this systems endianness.
// //////////////////////////////////////

#if NA_SYSTEM_ENDIANNESS == NA_ENDIANNESS_LITTLE

  // Native <-> Little. Single converters
  static NA_INLINE void naConvertNativeLittle128(void* buffer){
    naConvertNative128(buffer);
  }
  static NA_INLINE void naConvertNativeLittle64( void* buffer){
    naConvertNative64(buffer);
  }
  static NA_INLINE void naConvertNativeLittle32( void* buffer){
    naConvertNative32(buffer);
  }
  static NA_INLINE void naConvertNativeLittle16( void* buffer){
    naConvertNative16(buffer);
  }
  static NA_INLINE void naConvertNativeLittle8(  void* buffer){
    naConvertNative8(buffer);
  }

  // Native <-> Little. Array converters
  static NA_INLINE void naConvertArrayNativeLittle128(void* buffer,
                                                      NAInt count){
    naConvertArrayNative128(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeLittle64( void* buffer,
                                                      NAInt count){
    naConvertArrayNative64(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeLittle32( void* buffer,
                                                      NAInt count){
    naConvertArrayNative32(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeLittle16( void* buffer,
                                                      NAInt count){
    naConvertArrayNative16(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeLittle8(  void* buffer,
                                                      NAInt count){
    naConvertArrayNative8(buffer, count);
  }

  // Native <-> Big. Single converters
  static NA_INLINE void naConvertNativeBig128(void* buffer){
    naConvertLittleBig128(buffer);
  }
  static NA_INLINE void naConvertNativeBig64( void* buffer){
    naConvertLittleBig64(buffer);
  }
  static NA_INLINE void naConvertNativeBig32( void* buffer){
    naConvertLittleBig32(buffer);
  }
  static NA_INLINE void naConvertNativeBig16( void* buffer){
    naConvertLittleBig16(buffer);
  }
  static NA_INLINE void naConvertNativeBig8(  void* buffer){
    naConvertLittleBig8(buffer);
  }

  // Native <-> Big. Array converters
  static NA_INLINE void naConvertArrayNativeBig128(void* buffer,
                                                   NAInt count){
    naConvertArrayLittleBig128(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeBig64( void* buffer,
                                                   NAInt count){
    naConvertArrayLittleBig64(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeBig32( void* buffer,
                                                   NAInt count){
    naConvertArrayLittleBig32(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeBig16( void* buffer,
                                                   NAInt count){
    naConvertArrayLittleBig16(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeBig8(  void* buffer,
                                                   NAInt count){
    naConvertArrayLittleBig8(buffer, count);
  }

#elif NA_SYSTEM_ENDIANNESS == NA_ENDIANNESS_BIG

  // Native <-> Little. Single converters
  static NA_INLINE void naConvertNativeLittle128(void* buffer){
    naConvertLittleBig128(buffer);
  }
  static NA_INLINE void naConvertNativeLittle64( void* buffer){
    naConvertLittleBig64(buffer);
  }
  static NA_INLINE void naConvertNativeLittle32( void* buffer){
    naConvertLittleBig32(buffer);
  }
  static NA_INLINE void naConvertNativeLittle16( void* buffer){
    naConvertLittleBig16(buffer);
  }
  static NA_INLINE void naConvertNativeLittle8(  void* buffer){
    naConvertLittleBig8(buffer);
  }

  // Native <-> Little. Array converters
  static NA_INLINE void naConvertArrayNativeLittle128(void* buffer,
                                                      NAInt count){
    naConvertArrayLittleBig128(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeLittle64( void* buffer,
                                                      NAInt count){
    naConvertArrayLittleBig64(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeLittle32( void* buffer,
                                                      NAInt count){
    naConvertArrayLittleBig32(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeLittle16( void* buffer,
                                                      NAInt count){
    naConvertArrayLittleBig16(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeLittle8(  void* buffer,
                                                      NAInt count){
    naConvertArrayLittleBig8(buffer, count);
  }

  // Native <-> Big. Single converters
  static NA_INLINE void naConvertNativeBig128(void* buffer){
    naConvertN128(buffer);
  }
  static NA_INLINE void naConvertNativeBig64( void* buffer){
    naConvertN64(buffer);
  }
  static NA_INLINE void naConvertNativeBig32( void* buffer){
    naConvertN32(buffer);
  }
  static NA_INLINE void naConvertNativeBig16( void* buffer){
    naConvertN16(buffer);
  }
  static NA_INLINE void naConvertNativeBig8(  void* buffer){
    naConvertN8(buffer);
  }

  // Native <-> Big. Array converters
  static NA_INLINE void naConvertArrayNativeBig128(void* buffer,
                                                   NAInt count){
    naConvertArrayN128(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeBig64( void* buffer,
                                                   NAInt count){
    naConvertArrayN64(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeBig32( void* buffer,
                                                   NAInt count){
    naConvertArrayN32(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeBig16( void* buffer,
                                                   NAInt count){
    naConvertArrayN16(buffer, count);
  }
  static NA_INLINE void naConvertArrayNativeBig8(  void* buffer,
                                                   NAInt count){
    naConvertArrayN8(buffer, count);
  }

#else
  // Other cases are not implemented yet.
#endif






#endif // NA_BINARY_DATA_INCLUDED


// Copyright (c) NALib, Tobias Stamm, Manderim GmbH
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the source-code.
//
// In case the source-code of this software is inaccessible to the end-user,
// the above copyright notice and this permission notice shall be included
// in any source-code which is dependent on this software and is accessible
// to the end-user.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
