
// This file is part of NALib, a collection of C and C++ source code
// intended for didactical purposes. Full license notice at the bottom.


// In this file, basic functionality of pointers is implemented.
// NALib usually uses the native pointers of C and C++ but for some types, a
// special kind of pointer is more appropriate.
// 
// Other programming languages are based on pointers with reference counting
// which point to objects which are automatically allocated and deallocated
// by a clever runtime-system. In NALib, such pointers are only used where
// necessary. The native pointers are fine for most purposes.
//
// If somebody ever wants to write a runtime-system for NALib, this file
// would be a good place to start.


#ifndef NA_POINTER_INCLUDED
#define NA_POINTER_INCLUDED

#include "NAMathOperators.h"


// These functions are the allocation functions of NALib. All functions in NALib
// use these functions when allocating memory on the heap. The basic function
// simply allocates using malloc. The ifNull-function takes an additional prt
// parameter. If ptr is NULL, memory is allocated and the pointer to that
// memory is returned. If the pointer is not NULL, nothing is allocated, the
// same pointer is returned.
//
// Having one single allocation function helps detecting basic memory errors.
// Note however that there does not exist any exception handling in NALib,
// meaning an error might be detected though not resolved. And in favor of
// simplicity, NALib will not get an exception handling soon.
static NA_INLINE void* naAllocate(NAInt size){
  void* ptr; // Declaration before implementation.
  #ifndef NDEBUG
    if(size < 1)
      naError("naAllocate", "size is smaller than 1 .");
  #endif
  ptr = malloc(size);
  #ifndef NDEBUG
    if(!ptr)
      {naCrash("naAllocate", "out of memory."); return NA_NULL;}
  #endif
  return ptr;
}


static NA_INLINE void* naAllocateIfNull(void* ptr, NAInt size){
  #ifndef NDEBUG
    if(size < 1)
      naError("naAllocateIfNull", "size is smaller than 1 .");
  #endif
  if(!ptr){ptr = naAllocate(size);}
  return ptr;
}




// /////////////////////////////////////////
//
// The following is the inplementation of a pointer having a reference count,
// as used in many modern languages. In C and C++, this must be implemented
// manually.
//
// This implementation defines the ownership of the data using the sign of the
// refcount field.
//
// As having a reference count is not always useful in C and C++, only a few
// structs of NALib like the NAByteArray actually use the NAPointer structure.
//
// /////////////////////////////////////////




typedef struct NAPointer NAPointer;
struct NAPointer{
  void* data;           // The stored data (heap allocation).
  NAInt refcount;       // Reference count.
                        // When positive: This NAPointer owns the data
                        // When negative: This NAPointer does not own the data
                        // When 0: Pointer deletes itself. When this NAPointer
                        // owns the data, it deletes it as well.
};
// This whole struct should not be visible to the programmer. But to stay
// simple and to allow the compiler to inline code, it is plain visible.
//
// Try not to use any of the fields directly but use the functions below!



// Creates an NAPointer struct with a newly allocated memory block (heap)
// of the given size. The memory is uninitialized. The reference counter is
// automatically set to 1 meaning that this NAPointer is owning the memory
// and will delete it using free() when the reference count reaches 0.
static NA_INLINE NAPointer* naCreatePointerWithSize(NAInt size){
  NAPointer* pointer; // Declaration before implementation.
  #ifndef NDEBUG
    if(size < 1)
      naError("naCreatePointerWithSize", "size is smaller than 1 .");
  #endif
  pointer = naAllocate(sizeof(NAPointer));
  pointer->data = naAllocate(size);
  pointer->refcount = 1;
  return pointer;
}


// Creates an NAPointer struct with the pointer of the given buffer
// WITHOUT COPYING its contents. The given buffer can be located
// anywhere in memory. When this pointer takes the ownership, it will delete
// the memory using free() when the reference counter reaches 0.
static NA_INLINE NAPointer* naCreatePointerWithExistingBuffer(void* buffer,
                                                        NABool takeownership){
  NAPointer* pointer; // Declaration before implementation.
  #ifndef NDEBUG
    if(!buffer)
      naError("naCreatePointerWithExistingBuffer", "buffer is NULL.");
    if((takeownership != NA_FALSE) && (takeownership != NA_TRUE))
      naError("naCreatePointerWithExistingBuffer", "invalid ownership");
  #endif
  pointer = naAllocate(sizeof(NAPointer));
  pointer->data = buffer;
  pointer->refcount = (((NAInt)takeownership)<<1)-1;
  return pointer;
}


// Retains the given pointer. Meaning: There is one more codeblock which is
// using this NAPointer. This NAPointer will not be freed as long as that
// codeblock does not releases this NAPointer.
//
// Of course, someone else might erroneously release this NAPointer, but that's
// the risk you take with reference-counted pointers.
//
// Returns the pointer again. Makes it easier to write.
static NA_INLINE NAPointer* naRetainPointer(NAPointer* pointer){
  #ifndef NDEBUG
    if(!pointer)
      {naCrash("naRetainPointer", "pointer is NULL."); return NA_NULL;}
    // The next test can detect some erroneous behaviour in the code. Note
    // however that most likely the error did occur long before reaching here.
    if(pointer->refcount == 0)
      naError("naRetainPointer", "Retaining NAPointer with a refcount of 0");
  #endif
  pointer->refcount += naSigni(pointer->refcount);
  return pointer;
}


// Releases the given pointer. If the refcount reaches 0, this NAPointer
// is no longer needed and will be freed. If this NAPointer owns the data, it
// will be automatically freed as well.
static NA_INLINE void naReleasePointer(NAPointer* pointer){
  #ifndef NDEBUG
    if(!pointer)
      {naCrash("naReleasePointer", "pointer is NULL."); return;}
    // The next test can detect some erroneous behaviour in the code. Note
    // however that most likely the error did occur long before reaching here.
    if(pointer->refcount == 0)
      naError("naReleasePointer", "Releasing NAPointer with a refcount of 0");
  #endif
  if(pointer->refcount == 1){free(pointer->data);}
  // Note that the author decided to always count to zero, even if it is clear
  // from the previous line of code that the pointer will eventually be freed
  // and the data will be lost in nirvana. But often times in debugging, when
  // retaining and releasing is not done correctly, an NAPointer is released
  // too often. When refcount is 0 and NDEBUG is not defined, this can be
  // detected!
  pointer->refcount -= naSigni(pointer->refcount);
  // The following line is the reason why you can not initialize an existing
  // NAPointer struct with a createXXX functions: Because it will always be
  // freed automatically when refcount reaches 0.
  if(pointer->refcount == 0){free(pointer);}
  // Note that other programming languages have incorporated this very idea
  // of self-organized reference-counting pointers deeply within its core.
  // Their runtime-systems keep giant pools of free objects at hand and take
  // care of detecting and collecting unused objects. In C and C++, no such
  // mechanisms exist. You could write one though.
}


// Returns a pointer to the data. This function is not particularily beautiful
// when it comes to readability or writeability but it illustrates two things:
// - First, note that if the NAPointer struct would be hidden, there would be
//   no other way of accessing the data.
// - Second, this illustrates how accessors and mutators would be handeled in C
//   and C++: Simply create two functions where one returns a const pointer
//   (or a reference in C++), the other one returns a non-const pointer (or
//   reference). The first variant is an accessor, the second one a mutator.
// To stay simple, NALib will not implement Accessors and Mutators and not hide
// the struct declaration. For now.
static NA_INLINE void* naGetPointerData(NAPointer* pointer){
  #ifndef NDEBUG
    if(!pointer)
      {naCrash("naGetPointerData", "pointer is NULL."); return NA_NULL;}
  #endif
  return pointer->data;
}




// ////////////////////////
// Explanation: CONSTRUCTORS
// ////////////////////////

// Quite often, structs must be initialized before being able to use them.
// As NALib is a C library, there are no constructors like in typical OOP
// languages and consequently, such initialization must be done manually by
// the programmer.
//
// In NALib, initialization-functions always start with createXXX. They
// assume the given struct to be uninitialized and will overwrite all values.
// The struct itself can either be already allocated outside of the createXXX
// function or being automatically allocated within.
//
// 1. If the given first parameter of a createXXX function is NOT NULL, the
//    struct at the given pointer is used.
//    Outside of the function, the given struct is usually already stored in
//    a variable and a pointer to that variable is used in the function call.
//    For example:
//
//    NAByteArray myarray;
//    naCreateByteArrayWithSize(&myarray, 1234);
//
//    When no longer used, these structs usually are destructed with the
//    appropriate clearXXX function. See below.
//
//    To stay consistent with the second case, the pointer to the given struct
//    is returned at the end of the function.
//
// 2. If the given first parameter of a createXXX function is a NULL-pointer,
//    a new struct is allocated on the heap using malloc and at the end of the
//    function, the pointer to that memory location is returned.
//    Outside of the function, the returned pointer is usually stored in a
//    pointer variable. For example:
//
//    NAByteArray* myarray = naCreateByteArrayWithSize(NA_NULL, 1234);
//
//    When no longer used, these structs usually are deleted using the
//    appropriate destroyXXX function. See below.
//
//    Important notice: Using a NULL-pointer might seem to be the easiest way
//    when createing new structs. But each time, a new (likely small) space
//    must be allocated on the heap and later be freed again. In contrast to
//    modern programming languages, C is not optimized for this case and might
//    result in quite noticeable speed drops.
//
// Note that the NAPointer struct is an exception: Its createXXX functions will
// always allocate on the heap using method 2.



// ////////////////////////
// Explanation: DESTRUCTORS
// ////////////////////////

// Structs which have been initialized using a createXXX function (see above)
// must be deleted with the appropriate destructor.
//
// As NALib is a C library, there are no destructors like in typical OOP
// languages and consequently, such deletion must be done manually by
// the programmer.
//
// In NALib, destruction-functions either start with destroyXXX or clearXXX.
// Both functions expect a pointer to the struct and will perform all
// appropriate destructions. Basically, clearXXX will erase the contents.
// destroyXXX will call clearXXX first and free the given pointer after that.
//
// Often times, a programmer just wants to erase the contents of a variable
// but use it again immediately after that. Whereever this is useful, the
// clearXXX functions can have a second parameter which does just that.
//
// 1. clearXXX functions are meant for variables whose content shall be erased
//    but the memory of the struct itself shall not be freed.
//
//    A clearXXX function can sometimes have a second argument: reinitialize.
//    If this parameter is set to NA_TRUE, a createXXX function call will be
//    performed automatically. If the parameter is NA_FALSE, the values in the
//    struct are garbage.
//
// 2. destroyXXX functions are meant for Pointers which shall be erased and
//    freed completely from memory.
//
//    Do not use the destroyXXX functions on stack variables!
//
//
// Usually, the cases listed here go in pairs with the cases listed in the
// CONSTRUCTORS section above. Meaning a case 1 constructor usually will
// require a case 1 destructor and a case 2 constructor usually a case 2
// destructor. But in certain cases, a programmer might want to choose
// differently, for example to save time. If you choose so, just make sure
// you know what you are doing.
//



#endif // NA_POINTER_INCLUDED




// Copyright (c) NALib, Tobias Stamm, Manderim GmbH
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the source-code.
//
// In case the source-code of this software is inaccessible to the end-user,
// the above copyright notice and this permission notice shall be included
// in any source-code which is dependent on this software and is accessible
// to the end-user.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
