
// This file is part of NALib, a collection of C and C++ source code
// intended for didactical purposes. Full license notice at the bottom.

#ifndef NA_SYSTEM_INCLUDED
#define NA_SYSTEM_INCLUDED


// Types always have the prefix NA
// macros always have the prefix NA_ and are uppercase
// functions always have the prefix na
// variables always have the prefix na_ and are lowercase


// System dependant definitions
// ----------------------------
// The following will be defined after including this file:
//
// NA_SYSTEM                Denotes this system. Will be defined as one of
//                          the NA_SYSTEM_XXX below.
// NA_SYSTEM_ENDIANNESS     Denotes the endianness of this system. Will be
//                          defined as one of the NA_ENDIANNESS_XXX below.
// NA_SYSTEM_ADDRESS_BITS   Number of bits per address as an NAInt.
// NA_RESTRICT              The restrict keyword, but working cross-plattform.
// va_copy(d,s)             A macro which is missing on some systems.
//
// The following types and macros will be defined as well:
//
// int64, int32, int16, int8      Signed integers with exact number of bits.
// uint64, uint32, uint16, uint8  Unsigned integers with exact number of bits.
// NA_INT64_MAX, ...              Maximal values for all integer types.
// NA_INT64_MIN, ...              Minimal values for all signed integer types.
// NAByte                           Special name for an uint8.
// NAInt                          The most used int type of NALib.
//                                Is 32 Bits or 64 Bits depending on the
//                                architecture the code is build for.
// NABool, NA_TRUE, NA_FALSE      Boolean type and values.
//
// Additionally, there are some debugging definitions used throughout NALib.
// More informations below.
//
// naError                        A function emitting an error to stderr.
// naCrash                        A function emitting crashing with a message.

// The version of this very distribution. No sub-version or build number or
// whatever. Just one single number.
#define NA_LIB_VERSION 0

// The various Systems:
#define NA_SYSTEM_MAC_OS_X  0
#define NA_SYSTEM_WINDOWS   1
#define NA_SYSTEM_UNKNOWN   2
#define NA_SYSTEM_COUNT     3

// The various Endianness-Settings:
#define NA_ENDIANNESS_NATIVE  0
#define NA_ENDIANNESS_LITTLE  1
#define NA_ENDIANNESS_BIG     2
#define NA_ENDIANNESS_UNKNOWN 3 // Behaves the same as native.
#define NA_ENDIANNESS_COUNT   4
// Note that when expecting the endianness as a parameter of a function,
// it will have the type NAInt.


// ASCII-Strings for the macros above
extern const char* na_system_strings[NA_SYSTEM_COUNT];
extern const char* na_endianness_strings[NA_ENDIANNESS_COUNT];
extern const char* na_boolean_strings[2];



// //////////////////////////////////////////
// Try to figure out what system this is.
// //////////////////////////////////////////

// Interesting read: http://sourceforge.net/p/predef/wiki/OperatingSystems/
#if defined __APPLE__ && __MACH__
  #define NA_SYSTEM NA_SYSTEM_MAC_OS_X
  #if defined __LITTLE_ENDIAN__
    #define NA_SYSTEM_ENDIANNESS NA_ENDIANNESS_LITTLE
  #elif defined __BIG_ENDIAN__
    #define NA_SYSTEM_ENDIANNESS NA_ENDIANNESS_BIG
  #endif
  #if defined __LP64__
    #define NA_SYSTEM_ADDRESS_BITS 64
  #else
    #define NA_SYSTEM_ADDRESS_BITS 32
  #endif
  
#elif (defined _WIN64) || (defined _WIN32)
  #define NA_SYSTEM NA_SYSTEM_WINDOWS
  #define NA_SYSTEM_ENDIANNESS NA_ENDIANNESS_LITTLE
  #if defined _WIN64
    #define NA_SYSTEM_ADDRESS_BITS 64
  #else
    #define NA_SYSTEM_ADDRESS_BITS 32
  #endif

#else
  #warning "System not detected"
#endif

#ifndef NA_SYSTEM_ADDRESS_BITS
  #warning "Address bits not detected, trying 32."
  #define NA_SYSTEM_ADDRESS_BITS 32
#endif

#ifndef NA_SYSTEM_ENDIANNESS
  #warning "Endian format not detected, trying native."
  #define NA_SYSTEM_ENDIANNESS NA_ENDIANNESS_UNKNOWN
#endif



// ////////////////////////////////////////////////////////////
// System dependant mapping of global functions and macros
// ////////////////////////////////////////////////////////////

// A function declared with NA_HIDDEN will not be exported when building a
// library. This means that this function will not be listed in the .lib file
// on windows and will not be accessible anywhere when including the (.dll or
// .dylib) library. A function declared with NA_API will explicitely be
// exported. This is a system-independent implementation which allows the
// programmer to define the exporting on a very low granularity. There exist
// other methods to define the exporting in external files.
//
// Usually, the NA_HIDDEN macro is not visible to the programmer as it will
// only be used on hidden parts (seem legit, doesn't it?). Nonetheless it is
// listed here for better understanding.
//
// NA_API is just here for explanation. It is not used in NALib, as it will
// never be available as a library. Hence its name: NALib = Not A Library.

#if NA_SYSTEM == NA_SYSTEM_WINDOWS
  #define NA_HIDDEN
  #define NA_API      __declspec(dllexport)
  #define NA_RESTRICT __restrict
  #define NA_INLINE   _inline
  #define va_copy(d,s) ((d) = (s))
#elif NA_SYSTEM == NA_SYSTEM_MAC_OS_X
  #define NA_HIDDEN   __attribute__ ((visibility("hidden")))
  #define NA_API      __attribute__ ((visibility("default")))
  #define NA_RESTRICT __restrict__
  #define NA_INLINE   inline
#else
#endif




// /////////////////////////
// The integer types
// /////////////////////////

// The following basic integer types might also be defined using the typedefs
// provided in <stdint.h> or <cstdint>. As these files do not necessarily exist
// in all systems, the typedefs are provided differently in NALib.
//
// Note that these types do not have the NA prefix. The author decided that
// these types are too fundamental such that they do not need to be marked with
// the prefix. They should not collide with other typedefs found in other
// sourcecodes. If the do collide, there would be a serious problem and the
// implementation is likely incompatible. Please send the author an email if
// that happends.

#if NA_SYSTEM == NA_SYSTEM_WINDOWS
  typedef unsigned __int64  uint64;
  typedef unsigned __int32  uint32;
  typedef unsigned __int16  uint16;
  typedef unsigned __int8   uint8;
  typedef signed __int64    int64;
  typedef signed __int32    int32;
  typedef signed __int16    int16;
  typedef signed __int8     int8;
#elif NA_SYSTEM == NA_SYSTEM_MAC_OS_X
  #include <sys/types.h>
  typedef u_int64_t         uint64;
  typedef u_int32_t         uint32;
  typedef u_int16_t         uint16;
  typedef u_int8_t          uint8;
  typedef int64_t           int64;
  typedef int32_t           int32;
  typedef int16_t           int16;
  typedef int8_t            int8;
#else
#endif

// Note that the following macros might also be defined using the values
// provided in the <limits.h> or <climits> header file. In NALib, the are
// provided manually to not be dependent on the various definitions of
// char, short, int, long and long long.

#define NA_UINT8_MAX  ((uint8) (0xff))
#define NA_UINT16_MAX ((uint16)(0xffff))
#define NA_UINT32_MAX ((uint32)(0xffffffff))
#define NA_UINT64_MAX ((uint64)(0xffffffffffffffffLL))
// note that the - 1 is needed to avoid warnings on some compilers
#define NA_INT8_MAX   ((int8)  (0x7f))
#define NA_INT8_MIN   ((int8)  (0x81 - 1))
#define NA_INT16_MAX  ((int16) (0x7fff))
#define NA_INT16_MIN  ((int16) (0x8001 - 1))
#define NA_INT32_MAX  ((int32) (0x7fffffff))
#define NA_INT32_MIN  ((int32) (0x80000001 - 1))
#define NA_INT64_MAX  ((int64) (0x7fffffffffffffffLL))
#define NA_INT64_MIN  ((int64) (0x8000000000000001LL - 1LL))



// ////////////////////////////////////
// Other Basic types used in NALib
// ////////////////////////////////////

// A NAByte is a type definition of a NAByte. Like... you know... a NAByte.
// Seriously: Defining a NAByte as an unsigned char can be very handy.
// Further more, a Pointer to a NAByte can be displayed by a debugger while
// a pointer to void can not.
typedef uint8     NAByte;

// The NAInt type is an integer of the size which is needed for storing an
// address. This means that this type is dependent on the system NALib is
// compiled for. It can for example be 32 Bits on one and 64 Bits on another
// architecture.
//
// Note that the NAInt type is signed! This means that for example in
// 32-Bit systems, only about 2 billion entries can be accessed with arrays
// when using NALib. The author is perfectly aware of the fact that most
// standard library functions use an unsigned integer for size_t, and that
// NALib therefore will have some minor limitations regarding memory usage.
// But the author has been convinced by the simplicity and elegance of just
// using signed integers during the last couple of years. It makes programming
// and debugging so much easier. And in 64 Bit systems, the memory limitations
// will not be a problem for a couple of years coming.
//
// Note that integers with explicitely defined sizes are used. The definition
// of just "int" would be unreliable.
//
// One serious drawback of a system-dependent definition is that NALib might
// behave differently depending on the system it is compiled for. But it will
// only behave differently in cases where the behaviour would become difficult
// anyway. Meaning: You will most likely run into out-of-memory problems first.
//
// Note that a different solution would be to always go for int64 and convert
// between the different integer types. But compiling NALib on a non-64-Bit
// system then would definitely be a waste of space with lots of implicit
// arithmetic conversions.
//
#if NA_SYSTEM_ADDRESS_BITS == 32
  typedef int32 NAInt;
#elif NA_SYSTEM_ADDRESS_BITS == 64
  typedef int64 NAInt;
#else
  #warning "NAInt undefined"
#endif


// Note that in NALib, there is no base typedef for a void* or a const void*
// pointer. Some libraries or frameworks define them with names such as
// NAVoid, NAPtr, NABuffer, NABuf, NAIn, NAOut, NAInOut or such alike. The
// author believes that void* pointers should only be typedef'd if they
// describe a distinct purpose, like for example if a returned void* points
// to a group of structs sharing some common properties.
//
// If void* is used plainly as a pointer to an untyped address, the author
// thinks that it is best to just write void*, not some abscure type name.


// Note that in NALib, the definition of NABool is explicitely set to "int" and
// not char or unsigned char or not even NAInt. Just plain old "int".
// This is unusual but most probably the easiest way to tell the compiler to 
// use whatever he likes as long as it's fast. it is in the believe of the
// author, that in modern computers, speed is more important than space as
// opposed to earlier times where wasting space was a complete no-go. When
// using lots of Boolean values, there are better ways to store them than in
// NABools anyway: Use C-style-masks or have a look at NABitArray.
//
// Also note that there is no distinction between C and C++. Therefore, NALib
// never uses the bool type or the true and false keywords.
//
typedef int NABool;
#define NA_TRUE    1
#define NA_FALSE   0



// The definition of NA_NULL usually set to the NULL found in stdlib. The
// new C11 standard however has a new keyword. Let's use it if it is available!
// Note that stdlib is needed anyway for malloc free, exit.
//
// Also note that in the comments of NALib as well as in the error messages,
// it will always be called "NULL". In the implementation however, NA_NULL
// is used.

#include <stdlib.h>
#if defined __STDC_VERSION__
  #if __STDC_VERSION__ >= 201112L // This indicates the C11 standard
    #define NA_NULL nullptr
  #endif
#endif
#ifndef NA_NULL
  #define NA_NULL NULL
#endif


// /////////////////////////////////
// Debugging definitions:
//
// The folowing definitions will only be defined if NDEBUG is undefined.
// Many implementations of NALib will perform many tests which will slow down
// the performance more or less considerably. If NDEBUG is defined however,
// no tests are performed whatsoever.
//
// Note that all tests will emit errors on sterr in the following format:
// Function_name: Errormessage\n
// Therefore, a programmer can simply set a breakpoint in the denoted function
// and start debugging.
//
// Also note that in NALib, any code executed within NDEBUG will not alter any
// values. In other words: Except from speed differences and outputs to strerr,
// NALib will NOT behave differently depending on NDEBUG being defined or not. 
// /////////////////////////////////

#ifndef NDEBUG

  // The error printing method. Errors will be emitted to the stderr output.
  // When NDEBUG is defined, these functions are OBSOLETE!
  #ifdef __cplusplus
    #include <cstdio>
  #else
    #include <stdio.h>
  #endif
  // Prints an error. When this function gets called, the ongoing of the
  // application is undefined. Sometimes, the error might affect everything
  // which comes after it, including crashing the application, sometimes, the
  // error will just result in a NaN or even be corrected automatically.
  // Nontheless, any error should be considered a potential rist for the
  // application to crash.
  static NA_INLINE void naError(  const char* functionsymbol,
                                  const char* message){
    fprintf(stderr, "Error in %s: %s\n", functionsymbol, message);
  }
  // Prints a crash message and crashes the program.
  // This function is used when the application experiences a critical error
  // like dereferencing a NULL-Pointer. Even if NDEBUG would be defined, the
  // application will most likely crash somewhere after this function call.
  static NA_INLINE void naCrash(  const char* functionsymbol,
                                  const char* message){
    fprintf(stderr, "Critical Error in %s: %s\n", functionsymbol, message);
    fprintf(stderr, "The application will likely crash...\n");
    // Uncomment the following line, if you want to crash deliberately. By
    // default, this line is commented out as it would change the behaviour of
    // NALib depending on whether NDEBUG is defined or not.
    //
    // exit(1);
    //
    // Note that in the code, where naCrash is called, there always exists a
    // return instruction afterwards. This helps the compiler to determine that
    // a function can not crash, except deliberately.
  }

#endif



#endif // NA_SYSTEM_INCLUDED

// Copyright (c) NALib, Tobias Stamm, Manderim GmbH
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the source-code.
//
// In case the source-code of this software is inaccessible to the end-user,
// the above copyright notice and this permission notice shall be included
// in any source-code which is dependent on this software and is accessible
// to the end-user.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
